#!/usr/bin/env python
# -*- coding: latin1 -*-
#
##############################################################################
#
# metacomp 2.4beta1: a metacompiler for RLL(1) grammars
# Copyright (C) 2004 Juan Miguel Vilar and Andrs Marzal
#                    Universitat Jaume I, Castell (Spain)
#
# This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
#  Any questions regarding this software should be directed to:
#
#   Juan Miguel Vilar
#   Departament de Llenguatges i Sistemes Informtics
#   Universitat Jaume I
#   E12071 Castelln (SPAIN)
#
#   email: jvilar@lsi.uji.es
#
##############################################################################
#
# Fichero: gramatica.py
#
# $Date: 2004/04/26 14:58:33 $
#
# $Revision: 1.3 $
#
# $Log: gramatica.py,v $
# Revision 1.3  2004/04/26 14:58:33  jvilar
# Aadida la opcin de analizadores puros.
#
# Revision 1.2  2004/03/15 17:34:35  jvilar
# Actualizado a la versin 2.3.
#
#

import string

# Funciones auxiliares:

def _anyade(l, i): # Aade i a l si no est, devuelve 1 si lo aade
    if not i in l:
        l.append(i)
        return 1
    return 0

def _une(l1,l2): # Aade l2 a l1, devuelve 0 si l1 contiene a l2
    r= 0
    for i in l2:
        if not i in l1:
            r= 1
      	    l1.append(i)
    return r

def _interseccion(l1,l2): # Construye la interseccin de l1 y l2
    return filter(lambda x,y=l1: x in y, l2)

def _resta(l1, l2): # Elimina de l1 los elementos de l2
    return filter(lambda x,y=l2: x not in y, l1)

def _gestion_orden(orden, nombre):
    orden[nombre]= orden.get(nombre, 0)+1
    return `orden[nombre]`

def _nombre_lista(l):
    return string.join(map(lambda x: `x!= None and x.nombre or None`, l),", ")

def _str_lista(l):
    return string.join(map(lambda x: `x!= None and x.__str__() or None`, l),", ")

def _test_lista(l):
    if len(l)==1:
        return "mc_al.actual.cat== '%s'" % l[0]
    else:
        return "mc_al.actual.cat in [ %s ]" % ("'"+string.join(l,"', '")+"'")

def _test_no_lista(l):
    if len(l)==1:
        return "mc_al.actual.cat!= '%s'" % l[0]
    else:
        return "mc_al.actual.cat not in [ %s ]" % ("'"+string.join(l,"', '")+"'")

class Regla:
    "Representacin de las reglas"
    def __init__(self, izda, dcha, nl):
        self.nl= nl
        self.izda= izda
        self.dcha= dcha

    def anulable(self):
        return self.dcha.anulable()

    def __str__(self):
        return "%s -> %s ;" % (str(self.izda),str(self.dcha))

    def regla(self):
        return "%s -> %s ;" % (self.izda.regla(), self.dcha.regla())

    def aplica_funcion(self,f):
        self.izda.aplica_funcion(f)
        self.dcha.aplica_funcion(f)

class Pdcha:
    "Clase genrica para las partes derechas de las reglas"
    def __init__(self):
	pass

    def regla(self):
        # Devuelve la regla como cadena, sin las acciones
        return ""

    def anulable(self):
        # Devuelve 1 si es anulable, 0 si no lo es y None si no se sabe
        return None

    def simbolos(self):
        # Devuelve una lista con los terminales y no terminales
        return []

    def primeros(self):
        # Devuelve los primeros de parte derecha
        return self.misprimeros

    def calcula_primeros(self):
        # Calcula los primeros
        pass

    def aceptables(self):
        # Devuelve la unin de los primeros con los siguientes si es anulable
        return []

    def prepara_siguientes(self, pendientes, depende, inicial):
        # Prepara la bsqueda de los siguientes:
        #  pendientes: lista de nodos por tratar
        #  depende: lista de nodos de los que este depende para calcular
        #     los siguientes
        #  inicial: lista inicial de siguientes
        pass

    def prepara_aceptables(self):
        # Prepara los aceptables
        pass

    def genera_codigo(self, traza):
        # Genera el cdigo de la regla. Solo se usa para los no terminales
        #  traza: genera cdigo para escribir la traza (arbol o traza).
        return ""

    def genera_codigo_interno(self, pizda, orden, garantizados, traza, esPuro):
        # Genera la parte del cdigo de esta pdcha.
        #  pizda: parte izquierda de la regla que se est tratando
        #  ordenT: diccionarios para controlar la numeracin de terminales y
        #    no terminales.
        #  garantizados: lista de terminales que pueden llegar a este punto.
        #    Son siempre aceptables.
        #    None si no se sabe.
        #  traza: formato de la traza ("A" arbol, "t" traza).
        #  esPuro: cierto si se est generando un analizador sintctico puro.
        return ""

    def aplica_funcion(self,f):
        # Aplica la funcin f sobre s mismo y sus descendientes
        pass

    def recursividad_izquierda(self,nt):
        # Devuelve 1 si el nt tiene recursividad a izquierdas, 0 si no
        return 1


class NoTerminal(Pdcha):
    "Representa los no terminales"
    def __init__(self, nombre, nl):
        self.nl= nl
        self.nombre= nombre
        self.reglasizda= [] # reglas en las que este no terminal aparece
        		    # en la parte izquierda
        self.reglasdcha= [] # reglas en las que este no terminal aparece
        		    # en la parte derecha
        self.esanulable= None
	self.misprimeros= []
        self.missiguientes= []
        self.misaceptables= None
        self.influye=[] # nodos en los que influye para calcular los siguientes
        self.depende=[] # nodos de los que depende para calcular los siguientes
        self.traterror= None
        self.noterminales= [self]
        self.simbolos=[self]

    def anulable(self):
        return self.esanulable

    def __str__(self):
        return "<%s>" % self.nombre

    def regla(self):
        return str(self)

    def tratamientoError(self,tratamiento):
        self.traterror= tratamiento

    def siguientes(self):
        return self.missiguientes

    def aceptables(self):
        return self.misaceptables

    def prepara_siguientes(self, pendientes, depende, inicial):
        if depende:
            _anyade(self.depende, depende)
        _une(self.missiguientes, inicial)
        if _anyade(pendientes, self): # primera vez que pasamos por aqu
            for r in self.reglasizda:
                if r.dcha.prepara_siguientes(pendientes, self, self.missiguientes):
                    _anyade(self.influye, r.dcha)
        return 1

    def prepara_aceptables(self):
        if self.misaceptables== None:
            if self.esanulable:
                self.misaceptables= self.misprimeros[:]
                _une(self.misaceptables, self.missiguientes)
            else:
                self.misaceptables= self.misprimeros

    def genera_codigo(self, traza, esPuro= None):
        c= []
        if esPuro:
            c.append("def mc_analiza_%s(self):" % self.nombre)
        else:
            c.append("def mc_analiza_%s(self, %s):" % (self.nombre,self.nombre))
        indent="  "
        # La traza:
        if traza=="t":
            c.append(indent+"global indentacion_traza")
            c.append(indent+"""mc_traza("> %s %%s\\n" %% self.mc_al.actual)""" % self)
            c.append(indent+"indentacion_traza= indentacion_traza+2")
        elif traza=="A":
            c.append(indent+"global indentacion_traza")
            c.append(indent+r"""mc_traza('("%s" "lnea: %%d"\n' %% self.mc_al.actual.nlinea)""" % self)
            c.append(indent+"indentacion_traza= indentacion_traza+2")
        # Hacemos accesible el analizador lxico:
        c.append(indent+"mc_al= self.mc_al")
        # Comprobamos el tratamiento de errores:
        if self.traterror and not esPuro:
            c.append(indent+"self.mc_reintento.append(1)")
            c.append(indent+"mc_reintentar= self.mc_reintentar")
            c.append(indent+"while self.mc_reintento[-1]:")
            c.append(indent+"  self.mc_reintento[-1]= 0")
            c.append(indent+"  try:")
            indent="      "
        c.append(indent+"if not %s:" % _test_lista(self.misaceptables))
        if traza:
          c.append(indent+'  mc_traza("Error, no se encuentra ningn primero de %s.\\n")' % self)
        c.append(indent+"  mc_error('%s',%s)" % (self, `self.misaceptables` ))
        eselse= 0
        cadif= "if"
        n=0
        for r in self.reglasizda:
            n= n+1
            if len(self.reglasizda)> 1:
                if n== len(self.reglasizda):
                    c.append(indent+"else:")
                else:
                    c.append(indent+ cadif+ " %s:" % _test_lista(r.dcha.aceptables()))
                cadif="elif"
                indent= indent+"  "
            # La traza:
            if traza=="t":
                c.append(indent+"""mc_traza("Aplico regla: "+"""+ `r.regla()` +"""+"\\n")""")
            if not esPuro:
                # Generamos los atributos y los alias:
                orden= {}
                for s in r.dcha.simbolos:
                    _gestion_orden(orden, s.nombre)
                    if isinstance(s, NoTerminal):
                        nto= s.nombre+`orden[s.nombre]`
                        if orden[s.nombre]== 1 and s!= self:
                            c.append(indent+"%s= %s= Atributos()" % (s.nombre, nto))
                        else:
                            c.append(indent+"%s= Atributos()" % nto)
            # Generamos el cdigo de la regla:
            c2= r.dcha.genera_codigo_interno(self, {}, r.dcha.aceptables(), traza, esPuro)
            for i in c2:
                c.append(indent+i)
            if len(self.reglasizda)> 1:
                indent= indent[:-2]
        # Ms traza, para avisar de que nos vamos
        if traza=="t":
            c.append(indent+"""indentacion_traza = indentacion_traza-2""")
            c.append(indent+"""mc_traza("< %s\\n")""" % self)
        elif traza=="A":
            c.append(indent+"""indentacion_traza = indentacion_traza-2""")
            c.append(indent+"""mc_traza(")\\n")""")
        if self.traterror and not esPuro:
            c.append("    except mc_error_sintaxis, (mc_nt, mc_t):")
            indent= "      "
            for i in self.traterror:
                c.append(indent+i)
            c.append("  self.mc_reintento.pop()")
        return c

    def genera_codigo_interno(self, pizda, orden, garantizados, traza, esPuro):
        c=[]
        if not esPuro:
            atr= self.nombre+_gestion_orden(orden, self.nombre)
            c.append("self.mc_analiza_%s(%s)" % (self.nombre,atr))
            c.append("%s_= %s" % (self.nombre, atr))
        else:
            c.append("self.mc_analiza_%s()" % self.nombre)
        return c

    def aplica_funcion(self,f):
        f(self)

    def recursividad_izquierda(self, nt):
        return self==nt

    def comprueba_conflictos_globales(self, salida):
        vistos={}
        for r in self.reglasizda:
            for t in r.dcha.aceptables():
                if vistos.has_key(t):
                    salida.write("Aviso: hay un conflicto entre las regla de la lnea %d:\n"
                                 "    %s\n"
                                 "  y la de la lnea %d\n"
                                 "    %s\n"
                                 "  al ver el smbolo %s. Se tomar la primera.\n"
                                 % (vistos[t].nl, vistos[t].regla(), r.nl, r.regla(), t))
                else:
                    vistos[t]= r
            r.dcha.comprueba_conflictos(salida)

    def comprueba_conflictos(self, salida):
        pass

class Terminal(Pdcha):
    "Representa los terminales que no son la cadena vaca"
    def __init__(self, nombre, directo, error, nl):
        self.nombre= nombre
        self.directo= directo
        self.error= error
        self.nl= nl
        self.noterminales= []
        self.simbolos= [self]
        self.misprimeros= [self.nombre]

    def anulable(self):
        return 0

    def __str__(self):
        if self.directo:
            return '"%s"' % self.nombre[1:]
        else:
            return self.nombre

    def regla(self):
        return str(self)

    def aceptables(self):
        return [self.nombre]

    def prepara_siguientes(self, pendientes, depende, inicial):
        return 0

    def genera_codigo_interno(self, pizda, orden, garantizados, traza, esPuro ):
        c= []
        if garantizados and len(garantizados)==1 and garantizados[0]== self.nombre:
            indent=""
        else:
            indent="  "
            c.append("if mc_al.actual.cat == '%s':" % self.nombre)
        if traza=="t":
            c.append("""%smc_traza('Reconocido el terminal "%s"\\n')""" % (indent,self.nombre))
        elif traza=="A":
            c.append("""%smc_traza('("%s"\\n')""" % (indent,repr(string.replace(self.nombre,'"','""'))[1:-1]))
            c.append("""%smc_traza(' "Lexema: %%s"\\n' %% string.replace(mc_al.actual.lexema,'"','""'))""" % indent)
            c.append("""%smc_traza(' "Lnea: %%d"\\n' %% mc_al.actual.nlinea)""" % indent)
            c.append("""%smc_traza(')\\n')""" % indent) #'<- para el font-lock
        if not esPuro and not self.directo:
            n= self.nombre+_gestion_orden(orden, self.nombre)
            if orden[self.nombre]== 1 and pizda.nombre!= self.nombre:
                c.append("%s%s= %s= %s_= mc_al.actual" % (indent,self.nombre, n, self.nombre))
            else:
                c.append("%s%s= %s_= mc_al.actual" % (indent,n, self.nombre))
        c.append("%smc_al.avanza()" % indent)
        if traza=="t":
            c.append("""%smc_traza('Ledo el terminal "%%s"\\n' %% self.mc_al.actual)""" % indent)
        if indent=="  ":
            c.append("else:")
            if traza:
                c.append("""%smc_traza('Error no es un "%s"\\n')"""%(indent, self.nombre))
            if not self.error:
                c.append("  mc_error('%s', ['%s'])" % (pizda, self.nombre))
            else:
                for l in self.error:
                    c.append("  "+l)
        return c

    def aplica_funcion(self,f):
        f(self)

    def recursividad_izquierda(self,nt):
        return 0

    def comprueba_conflictos(self,salida):
        pass

class Vacia(Terminal):
    "Representa el terminal cadena vaca"
    def __init__(self, nl):
        self.noterminales=[]
        self.simbolos=[]
        self.misprimeros=[]
        self.nl= nl

    def anulable(self):
        return 1

    def __str__(self):
        return ""

    def regla(self):
        return ""

    def aceptables(self):
        return self.missiguientes

    def prepara_siguientes(self, pendientes, depende, inicial):
        if depende:
            self.depende= [depende]
        else:
            self.depende= []
        self.influye= []
        self.missiguientes=inicial[:]
        return 1

    def genera_codigo_interno(self, pizda, orden, garantizado, traza, esPuro):
        c= []
        if traza=="t":
            c.append("mc_traza('Reconocida una cadena vaca\\n')")
        elif traza=="A":
            c.append("""mc_traza('("")\\n')""")
        else:
            c.append("pass")
        return c

    def aplica_funcion(self,f):
        f(self)

    def recursividad_izquierda(self,nt):
        return 0

    def comprueba_conflictos(self,salida):
        pass

class Accion(Vacia):
    "Accin en una regla"
    def __init__(self, cod, nl):
        self.cod= cod
        self.noterminales=[]
        self.simbolos=[]
        self.misprimeros=[]
        self.nl= nl

    def __str__(self):
        return "@%s@" % self.cod

    def regla(self):
        return ""

    def genera_codigo_interno(self, pizda, orden, garantizado, traza, esPuro):
        if esPuro:
            return ["pass"]
        else:
            return [self.cod]

class Secuencia(Pdcha):
    "Parte derecha consistente en una secuencia de elementos"
    def __init__(self, l, nl):
        self.l= l
        self.nl= nl
        self.misnoterminales= None
        self.esanulable= None
        self.misprimeros= None
        self.noterminales= []
        self.simbolos= []
        for i in l:
            self.noterminales= self.noterminales+i.noterminales
            self.simbolos= self.simbolos+i.simbolos

    def anulable(self):
        if self.esanulable== None:
            for pd in self.l:
                aux= pd.anulable()
                if aux!= 1:     # Puede fallar por ser None o 0 :
                                #   Si es por ser cero, estamos seguros de que
                                # la secuencia no es anulable. En otro caso,
                                # seguimos sin saber.
                    if aux== 0:
                        self.esanulable= 0
                    break
            else: # Si todos son anulables, la secuencia lo es
                self.esanulable=1
        return self.esanulable

    def __str__(self):
        r=[]
        for i in self.l:
            r.append(str(i))
        return string.join(r)

    def regla(self):
        r=[]
        for i in self.l:
            if i.regla():
                r.append(i.regla())
        return string.join(r)

    def calcula_primeros(self):
        pr= []
        almacena= 1
        for pd in self.l:
            pd.calcula_primeros()
            if almacena:
                _une(pr, pd.primeros())
                if not pd.anulable():
                    almacena= 0
        self.misprimeros= pr

    def aceptables(self):
        return self.misaceptables

    def prepara_siguientes(self, pendientes, depende, inicial):
        if depende:
            self.depende=[depende]
        else:
            self.depende=[]
        self.missiguientes= inicial[:]
        self.influye=[]
        dep= self
        ini= self.missiguientes[:]
        r= 0
        for i in range(len(self.l)-1,-1,-1):
            pd= self.l[i]
            if pd.prepara_siguientes(pendientes, dep, ini) and dep:
                _anyade(self.influye, pd)
                r= 1
            if pd.anulable():
                _une(ini, pd.primeros())
            else:
                ini= pd.primeros()[:]
                dep= None
        return r

    def prepara_aceptables(self):
        if self.anulable():
            self.misaceptables= self.misprimeros[:]
            _une(self.misaceptables, self.missiguientes)
        else:
            self.misaceptables= self.misprimeros
        for pd in self.l:
            pd.prepara_aceptables()

    def genera_codigo_interno(self, pizda, orden, garantizados, traza, esPuro):
        c=[]
        g= garantizados
        if traza=="t":
            c.append("""mc_traza('> Analizando %s\\n')""" % self.regla())
            c.append("""indentacion_traza= indentacion_traza+2""")
        for i in self.l:
            c= c+ i.genera_codigo_interno(pizda, orden, g, traza, esPuro)
            if not isinstance(i,Vacia):
                g= None
        if traza=="t":
            c.append("""indentacion_traza= indentacion_traza-2""")
            c.append("""mc_traza('< Fin de %s\\n')""" % self.regla())
        return c

    def aplica_funcion(self,f):
        f(self)
        for i in self.l:
            i.aplica_funcion(f)

    def recursividad_izquierda(self,nt):
        for i in self.l:
            if i.recursividad_izquierda(nt):
                return 1
            if not i.anulable():
                return 0
        return 0

    def comprueba_conflictos(self,salida):
        for i in self.l:
            i.comprueba_conflictos(salida)

class Opcional(Pdcha):
    "Parte derecha representando la aparicin o no de una parte"

    def __init__(self, pd, nl):
        self.pd= pd
        self.nl= nl
        self.noterminales= pd.noterminales
        self.simbolos= pd.simbolos

    def anulable(self):
        return 1

    def __str__(self):
        return "( %s )?" % str(self.pd)

    def regla(self):
        return "( %s )?" % self.pd.regla()

    def calcula_primeros(self):
        self.pd.calcula_primeros()
        self.misprimeros= self.pd.misprimeros

    def aceptables(self):
        return self.misaceptables

    def prepara_siguientes(self, pendientes, depende, inicial):
        if depende:
            self.depende= [depende]
        else:
            self.depende= []
        self.missiguientes= inicial[:]
        if self.pd.prepara_siguientes(pendientes, self, inicial):
            self.influye= [self.pd]
        else:
            self.influye= []
        return 1

    def prepara_aceptables(self):
        self.misaceptables= self.primeros()[:]
        _une(self.misaceptables, self.missiguientes)
        self.pd.prepara_aceptables()

    def genera_codigo_interno(self, pizda, orden, garantizados, traza, esPuro):
        c=[]
        g= self.pd.primeros()
        if traza=="t":
            c.append("""mc_traza('> Analizando %s\\n')""" % self.regla())
        c.append("if %s:" % _test_lista(g))
        for l in self.pd.genera_codigo_interno(pizda, orden, g, traza, esPuro):
            c.append("  "+l)
        c.append("elif %s:" % _test_no_lista(self.missiguientes))
        if traza:
            c.append("  mc_traza('Error, no podemos asumir que est vaca.\\n')")
        c.append("  mc_error('%s', %s)" % (pizda, `self.misaceptables`))
        if traza=="t":
            c.append("else:")
            c.append("  mc_traza('Se supone que no est\\n')")  
        return c

    def aplica_funcion(self,f):
        f(self)
        self.pd.aplica_funcion(f)

    def recursividad_izquierda(self,nt):
        return self.pd.recursividad_izquierda(nt)

    def comprueba_conflictos(self, salida):
        for t in self.primeros():
            if t in self.missiguientes:
                salida.write("Aviso: hay un conflicto en la parte opcional de la lnea %d:\n"
                             "   %s\n"
                             "  con el smbolo %s. En ese caso, desplazar.\n"
                             % (self.nl, self.regla(), t))
        self.pd.comprueba_conflictos(salida)

class Disyuncion(Pdcha):
    "Parte derecha para representar multiples opciones"
    def __init__(self, l, nl, traterror= None):
        self.l= l
        self.nl= nl
        self.traterror= traterror
        self.esanulable= None
        self.noterminales= []
        self.simbolos= []
        for i in self.l:
            self.noterminales= self.noterminales+i.noterminales
            self.simbolos= self.simbolos+i.noterminales

    def anulable(self):
        if self.esanulable== None:
            a= 0
            for pd in self.l:
                if pd.anulable()== None:
                    a= None
                elif pd.anulable()== 1:
                    self.esanulable= 1
                    break
            else:
                self.esanulable= a
        return self.esanulable

    def __str__(self):
        r=[]
        for i in self.l:
            r.append(str(i))
        return "( %s )" % string.join(r," | ")

    def regla(self):
        r=[]
        for i in self.l:
            r.append(i.regla())
        return "( %s )" % string.join(r," | ")

    def calcula_primeros(self):
        pr= []
        for i in self.l:
            i.calcula_primeros()
            _une(pr, i.primeros())
        self.misprimeros= pr

    def aceptables(self):
        return self.misaceptables

    def prepara_siguientes(self, pendientes, depende, inicial):
        if depende:
            self.depende= [depende]
        else:
            self.depende= []
        self.missiguientes= inicial[:]
        self.influye=[]
        r= 0
        for pd in self.l:
            if pd.prepara_siguientes(pendientes, self, inicial):
                _anyade(self.influye, pd)
                r= 1
        return r

    def prepara_aceptables(self):
        self.misaceptables= []
        for pd in self.l:
            pd.prepara_aceptables()
            _une(self.misaceptables, pd.aceptables())

    def genera_codigo_interno(self, pizda, orden, garantizados, traza, esPuro ):
        c=[]
        el=""
        indent= ""
        if self.traterror and not esPuro:
            c.append("""self.mc_reintento.append(1)""")
            c.append("""mc_reintentar= self.mc_reintentar""")
            c.append("""while self.mc_reintento[-1]:""")
            c.append("""  self.mc_reintento[-1]= 0""")
            c.append("""  try:""")
            indent="    "
        if traza=="t":
            c.append(indent+"""mc_traza('> Analizando %s\\n')""" % self.regla())
            c.append(indent+"""indentacion_traza=indentacion_traza+2""")
        # No podemos hacer caso de las garantas si hay tratamiento de error;
        # no sabemos dnde se puede haber sincronizado.
        if garantizados and (esPuro or not self.traterror):
            g= garantizados[:]
        else:
            g= None
        for i in self.l:
            if g!= None:
                r= _interseccion(i.aceptables(), g)
            else:
                r= i.aceptables()
            if r:
                if g== None or len(r)< len(g): # Caso normal: los aceptables son menos que los garantizados
                    c.append(indent+el+"if %s:" % _test_lista(r))
                    el="el"
                    indent2=indent+"  "
                elif len(r)== len(g):
                    if el== "": # Esta alternativa se lo lleva todo y no hemos visto ninguna otra
                        indent2=indent+""
                    else: # Estamos en la ltima alternativa
                        indent2=indent+"  "
                        c.append("else:")
                for l in i.genera_codigo_interno(pizda, orden, r, traza, esPuro):
                    c.append(indent2+l)
            if g:
                g= _resta(g, r)
        if g== None: # Puede que haya un error en la entrada
            c.append(indent+"else:")
            if traza:
                c.append(indent+"  mc_traza('Error, no hay ninguna alternativa aceptable.\\n')")
            c.append(indent+"  mc_error('%s', %s)" % (pizda, `self.misaceptables`))
        if traza=="t":
            c.append(indent+"""indentacion_traza=indentacion_traza-2""")
            c.append(indent+"""mc_traza('< Fin de %s\\n')""" % self.regla())
        if self.traterror and not esPuro:
            c.append("  except mc_error_sintaxis, (mc_nt, mc_t):")
            indent= "    "
            for i in self.traterror:
                c.append(indent+i)
            c.append("self.mc_reintento.pop()")
        return c

    def aplica_funcion(self,f):
        f(self)
        for i in self.l:
            i.aplica_funcion(f)

    def recursividad_izquierda(self,nt):
        for i in self.l:
            if i.recursividad_izquierda(nt):
                return 1
        return 0

    def comprueba_conflictos(self, salida):
        vistos={}
        for r in self.l:
            for t in r.aceptables():
                if vistos.has_key(t):
                    salida.write("Aviso: hay un conflicto entre las opciones:\n"
                                 "    %s\n"
                                 "  y\n"
                                 "    %s\n"
                                 "  de la disyuncin de la lnea %d\n"
                                 "    %s\n"
                                 "  al ver el smbolo %s. Se tomar la primera.\n"
                                 % (vistos[t].regla(), r.regla(), self.nl, self.regla(), t))
                else:
                    vistos[t]= r
            r.comprueba_conflictos(salida)

class Iteracion(Pdcha):
    "Parte derecha representando cero o ms iteraciones"
    def __init__(self, pd, nl):
        self.pd= pd
        self.nl= nl
        self.noterminales= pd.noterminales
        self.simbolos= pd.simbolos

    def anulable(self):
        return 1

    def __str__(self):
        return "( %s )*" % str(self.pd)

    def regla(self):
        return "( %s )*" % self.pd.regla()

    def calcula_primeros(self):
        self.pd.calcula_primeros()
        self.misprimeros= self.pd.misprimeros

    def aceptables(self):
        return self.misaceptables

    def prepara_siguientes(self, pendientes, depende, inicial):
        if depende:
            self.depende= [depende]
        else:
            self.depende= []
        self.missiguientes= inicial[:]
        aux= inicial[:]
        _une(aux, self.primeros())
        if self.pd.prepara_siguientes(pendientes, self, aux):
            self.influye= [self.pd]
        else:
            self.influye= []
        return 1

    def prepara_aceptables(self):
        self.misaceptables= self.pd.primeros()[:]
        _une(self.misaceptables, self.missiguientes)
        self.pd.prepara_aceptables()

    def genera_codigo_interno(self, pizda, orden, garantizados, traza, esPuro):
        c=[]
        if traza=="t":
            c.append("""mc_traza('> Analizando %s\\n')""" % self.regla())
            c.append("""indentacion_traza= indentacion_traza+2""")
        c.append("while %s:" % _test_lista(self.pd.primeros()))
        for l in self.pd.genera_codigo_interno(pizda, orden, self.pd.primeros(), traza, esPuro):
            c.append("  "+l)
        if traza=="t":
            c.append("""indentacion_traza= indentacion_traza-2""")
            c.append("""mc_traza('< Fin de %s\\n')""" % self.regla())
        c.append("if not %s:" % _test_lista(self.missiguientes))
        if traza:
            c.append("  mc_traza('Error, no encontramos ninguno de los siguientes\\n')")
        c.append("  mc_error('%s',%s)" % (self.mi_nt, `self.misaceptables`))
        return c

    def aplica_funcion(self,f):
        f(self)
        self.pd.aplica_funcion(f)

    def recursividad_izquierda(self,nt):
        return self.pd.recursividad_izquierda(nt)

    def comprueba_conflictos(self, salida):
        for t in self.primeros():
            if t in self.missiguientes:
                salida.write("Aviso: hay un conflicto en la clausura de la lnea %d:\n"
                             "   %s\n"
                             "  con el smbolo %s. En ese caso, desplazar.\n"
                             % (self.nl, self.regla(), t))
        self.pd.comprueba_conflictos(salida)

#
#  Clase para las gramticas:
#

class Gramatica:
    def __init__(self, noterminales, inicial, reglas, noEOF):
        self.noEOF= noEOF
        self.noterminales= noterminales
	self.inicial= inicial
        self.reglas= reglas
	for r in reglas:
            r.izda.reglasizda.append(r)
	    nts= r.dcha.noterminales
	    for nt in nts:
	        if not r in nt.reglasdcha:
	            nt.reglasdcha.append(r)
        self.calcula_anulables()
        self.calcula_primeros()
        self.calcula_siguientes()
        self.prepara_aceptables()

        for r in self.reglas:
            def escribe_mi_nt(self, mi_nt= r.izda):
                self.mi_nt= mi_nt
            r.dcha.aplica_funcion(escribe_mi_nt)

    def __str__(self):
        s= []
        for r in self.inicial.reglasizda:
            s.append(str(r))
        for nt in self.noterminales:
            if nt!= self.inicial:
                for r in nt.reglasizda:
                    s.append(str(r))
        return string.join(s,"\n")

    def lista_reglas(self):
        l=[]
        for r in self.reglas:
            l.append(r.regla())
        return string.join(l,"\n")

    def calcula_anulables(self):
        pendientes=[]
    	for r in self.reglas:
           if r.dcha.anulable():
           	if r.izda.esanulable==None:
                    r.izda.esanulable=1
                    pendientes.append(r.izda)

	while pendientes:
	    nt= pendientes.pop()
	    for r in nt.reglasdcha:
	        if r.dcha.anulable():
	            if r.izda.esanulable==None:
	                r.izda.esanulable=1
	                pendientes.append(r.izda)

	for nt in self.noterminales:
	    if not nt.esanulable:
	        nt.esanulable= 0

    def calcula_primeros(self):
        l= self.reglas[:]
        while l:
            r= l.pop()
            r.dcha.calcula_primeros()
            if _une(r.izda.misprimeros, r.dcha.primeros()):
                _une(l,r.izda.reglasdcha)

    def calcula_siguientes(self):
        pendientes=[]
        for nt in self.noterminales:
            if nt!= self.inicial:
                nt.prepara_siguientes(pendientes, None, [])
            else:
                nt.prepara_siguientes(pendientes, None, ["mc_EOF"])
#          def escribe_influye(self):
#              try:
#                  print "%s influye en %s" % (self,_str_lista(self.influye))
#              except:
#                  print "%s no influye" % self
#          self.aplica_funcion(escribe_influye)

#          def escribe_depende(self):
#              try:
#                  print "%s depende de %s" % (self,_str_lista(self.depende))
#              except:
#                  print "%s no depende" % self
#          self.aplica_funcion(escribe_depende)
        while pendientes:
            nodo= pendientes.pop()
#            print "procesando %s" % nodo
            for nodo2 in nodo.depende:
                _une(nodo.missiguientes, nodo2.missiguientes)
#            print "   siguientes %s" % nodo.missiguientes
            for nodo2 in nodo.influye:
#                print "   influye en %s" % nodo2
#                print "       siguientes %s" % nodo2.missiguientes
                if _une(nodo2.missiguientes, nodo.missiguientes):
                    _anyade(pendientes, nodo2)
#                print "       actualizados %s" % nodo2.missiguientes

    def prepara_aceptables(self):
        for nt in self.noterminales:
            nt.prepara_aceptables()
        for r in self.reglas:
            r.dcha.prepara_aceptables()

    def aplica_funcion(self,f):
        for r in self.reglas:
            r.aplica_funcion(f)

    def recursividad_izquierda(self):
        l= []
        for nt in self.noterminales:
            for r in nt.reglasizda:
                if r.dcha.recursividad_izquierda(nt):
                    _anyade(l,nt)
        return l

    def comprueba_conflictos(self, salida):
        for nt in self.noterminales:
            nt.comprueba_conflictos_globales(salida)

def main():
  mas= Terminal("mas")
  menos= Terminal("menos")
  por= Terminal("por")
  entre= Terminal("entre")
  id= Terminal("id")
  num= Terminal("num")
  abre= Terminal("abre")
  cierra= Terminal("cierra")
  pyc= Terminal("pyc")

  terminales= [mas,menos,por,entre,id,num,abre,cierra]

  Expresion= NoTerminal("Expresion")
  Termino= NoTerminal("Termino")
  Factor= NoTerminal("Factor")
  Programa= NoTerminal("Programa")

  noterminales= [Programa,Expresion,Termino,Factor] 

  R0= Regla(Programa, Iteracion(Secuencia([Opcional(Expresion), pyc])));
  R1= Regla(Expresion, Secuencia([Termino,Iteracion(Secuencia([Disyuncion([mas,menos]),Termino]))]))
  R2= Regla(Termino, Secuencia([Factor,Iteracion(Secuencia([Disyuncion([por,entre]),Factor]))]))
  R3= Regla(Factor, Disyuncion([id,num,Secuencia([abre,Expresion,cierra])]))

  reglas= [R0, R1, R2, R3]

  g= Gramatica(noterminales, Programa, reglas)

  print g
  print "\n Smbolos anulables:"
  for i in noterminales:
      print i,i.anulable()

  print "\nPrimeros:"
  for nt in noterminales:
      l= []
      for i in nt.primeros():
          l.append(str(i))
      print "Pr(%s)= { %s }" % (nt, string.join(l,","))

  print "\nSiguientes:"
  for nt in noterminales:
      l= []
      for i in nt.siguientes():
          l.append(str(i))
      print "Sig(%s)= { %s }" % (nt, string.join(l,","))

if __name__ == "__main__":
    main()

